/*===========================================================================



Twin SBM-20 Geiger counter
==========================



Overview:
=========

CPU         :   Arduino Nano / ATmega328
Clock       :   XTAL 16MHz
Compiler    :   Arduino 1.0.5
Version     :   1.00
Date        :   2017-01-05
Author      :   Iacopo Giangrandi HB9DUL



Description:
============

This program implements the counting functions and unit conversion of a 
twin tube SBM-20 based Geiger counter. It accepts pulses on D2 and D3 
digital inputs. Only these two pins can be used, because they directly 
generates interrupt on change INT0 and INT1 respectively. Every time a 
pulse (falling edge) is detected, the corresponding counter is incremented 
by one. This counter is designed to handle two tubes for increased low 
radiation sensitivity. It will automatically correct the results by 
dividing the collected pulses in half which results in averaging out both 
inputs. Should only one tube be installed, shorting D19/A5 to GND will 
remove this averaging function and the software will work perfectly with 
just one tube. It doesn’t matter which input is used, only the division by 
two is removed. D19/A5 is only checked at the initialization and cannot be 
changed during runtime (it makes no sense!).

Results are displayed on an 16x2 LCD screen and the user can use three 
buttons to control the unit. Button P1 (unit) is connected to D16/A2 and 
allows changing the unit of the result between uR/h, nSv/h and cpm. Button 
P3 (time), connected to D18/A4, allows changing the integration time 
between 2s, 10s and 60s. Longer intervals give more accurate readings but 
take longer. A countdown on the top right corner of the display indicates 
the remaining time before the end of the integration. Button P2 (mode), 
connected to D17/A3, allows changing the mode of the instrument between 
maximum, minimum, individual tube counts and battery voltage. Pressing and 
holding P2 will reset the statistics.

Battery voltage is sampled via a voltage divider on pin D15/A1 and Vdd 
rail voltage is sampled via an identical divider on pin D14/A0. A diode 
connects the battery to the Vdd rail, while the power can also come from 
the Arduino USB connector, justifying the two different voltages.

Each time an integration period ends, the display is updated and the raw 
measurements are also echoed on the serial port at 9600bps. No calculation 
in mR/h nor uSv/h is done on the serial communication, as this can easily 
be done by the remote PC, based on the raw data. Tube conversion constants 
and data format is transmitted at boot time. As a reminder, the SBM-20 
tube produces 175cpm at 1uSv/h and 1530cpm at 1mR/h. Dead time is 190us at 
400V.



I/O connections:
================

LCD connections:
----------------
D8     ----> LCD_RS     LCD display Register Select line (0=command, 1=data)
D9     ----> LCD_EN     LCD display Enable line (1=enable)
GND    ----- LCD_R/W    LCD display Read/Write line (0=write)
D4     ----> LCD_D4     LCD display data bit 4
D5     ----> LCD_D5     LCD display data bit 5
D6     ----> LCD_D6     LCD display data bit 6
D7     ----> LCD_D7     LCD display data bit 7

UART connections:
-----------------
D0/RXD <---- USB_TX     UART communication (from PC to Arduino)
D1/TXD ----> USB_RX     UART communication (from Arduino to PC)

I/O connections:
----------------
D2     <---- IN_TUBE_1
D3     <---- IN_TUBE_2
D14/A0 <---- VDD voltage (analog)
D15/A1 <---- Battery voltage (analog)
D16/A2 <---- IN_P1 (UNIT)
D17/A3 <---- IN_P2 (MODE)
D18/A4 <---- IN_P3 (TIME)
D19/A5 <---- IN_SINGLE_TUBE



============================================================================= */



// --------------------------------------------------------------------------
// External files inclusions.
// --------------------------------------------------------------------------
#include <LiquidCrystal.h>                                      // Include dot-matrix LCD library.



// --------------------------------------------------------------------------
// General definitions, constants and global variables.
// --------------------------------------------------------------------------
#define       WELCOME_MSG_1     "Twin tube Geiger"              // Message shown on the display each time the device is powered up (line 1).
#define       WELCOME_MSG_2     "I.G.  2017  v1.0"              // Message shown on the display each time the device is powered up (line 2).

unsigned long int count_tube_1 = 0;                             // This variable is used to count pulses from tube 1. It's incremented by interrupt procedure int_tube_1().
unsigned long int count_tube_2 = 0;                             // This variable is used to count pulses from tube 2. It's incremented by interrupt procedure int_tube_2().
unsigned long int next_millis;                                  // This variable is used to calculate the expiration time of the current integration period.
long int cpm_current = 0;                                       // This variable contains the CPM (counts per minute) of the last completed (current) measurement.
long int cpm_max = 0;                                           // This contains the maximum of all measurements.
long int cpm_min = -1;                                          // This contains the minimum of all measurements (negative when reset).
long int cpm_avg = -1;                                          // This contains the average of all measurements (negative when reset).
unsigned long int avg_total_count;                              // Total counts for average calculations.
unsigned long int avg_total_millis;                             // Total duration in ms fo average calculation.
#define       AVG_RESET         0                               // In this status the average will be initialized at the next measurement.
#define       AVG_RUNNING       1                               // In this status the average is computed at each measurement.
#define       AVG_OVERFLOW      2                               // In this status the computation of the average is suspended because variables may overflow.
byte avg_status = AVG_RESET;                                    // Status of the average. 

#define       MODE_COOUNT       0                               // In this mode, the individual counts for each tube are displayed in real time in the bottom line.
#define       MODE_MAX          1                               // In this mode, the maximum reading is displayed in the bottom line.
#define       MODE_AVG          2                               // In this mode, the average reading is displayed in the bottom line.
#define       MODE_MIN          3                               // In this mode, the minimum reading is displayed in the bottom line.
#define       MODE_VOLT         4                               // In this mode, the voltages are displayed in real time in the bottom line.
char mode = MODE_COOUNT;                                        // This variable defnes the current mode. Here we deine the default mode when the device is powered on.

#define       UNIT_CPM          0                               // With this unit, counts per minute (cpm) are displayed.
#define       UNIT_R            1                               // With this unit, Roentgens per hour (R/h) are displayed.
#define       UNIT_SV           2                               // With this unit, Sieverts per hour (Sv/h) are displayed.
char unit = UNIT_CPM;                                           // This variable defnes the current unit. Here we deine the default unit when the device is powered on.

#define       TUBE_R_FACTOR     1530                            // 1530 cpm are equivalent to 1mR/h (SBM-20 specific data).
#define       TUBE_SV_FACTOR    175                             // 175 cpm are equivalent to 1uSv/h (SBM-20 specific data).
#define       TUBE_DEAD_TIME    190                             // The SBM-20 tube has a dead time of 190us (SBM-20 specific data).
byte num_tubes = 2;                                             // The number of installed tubes. Initially set to 2 and set to 1 at the initialization if IN_SINGLE_TUBE (D19/A5) is grounded.

#define       TIME_SLOW         60000                           // Long integration time is set to 60 seconds (60000ms).
#define       TIME_MED          10000                           // Medium integration time set to 10 seconds (10000ms).
#define       TIME_FAST         2000                            // Short long integration time set to 2 seconds (2000ms).
unsigned long time_interval = TIME_MED;                         // This variable defnes the current integration time. Here we deine the default integration time when the device is powered on.



// --------------------------------------------------------------------------
// Hardware related definitions, constants and global variables.
// --------------------------------------------------------------------------
#define       IN_TUBE_1         2                               // Input for pulses of tube 1. Must be on D2 to use INT0. Triggers on falling edge.
#define       IN_TUBE_2         3                               // Input for pulses of tube 2. Must be on D3 to use INT1. Triggers on falling edge.
#define       IN_P1             16                              // Input for button P1 (UNIT). LOW=pressed. Uses internal pull-up.
#define       IN_P2             17                              // Input for button P2 (MODE). LOW=pressed. Uses internal pull-up.
#define       IN_P3             18                              // Input for button P1 (TIME). LOW=pressed. Uses internal pull-up.
#define       IN_SINGLE_TUBE    19                              // Input for single tube jumper. LOW=single tube. Uses internal pull-up.
#define       VOLT_VDD          14                              // Analog input for VDD rail (through voltage divider).
#define       VOLT_BATT         15                              // Analog input for battery rail (through voltage divider).

#define       VDD_MIN           4500                            // If Vdd is below 4.5V, we definitely have a voltage problem, so a low battery warining is produced.
#define       VBATT_MIN         2900                            // The LiPo battery is conidered low if below 2.9V
#define       VBATT_ZERO        100                             // But if the battery voltage is below 0.1V, we have no battery and we are running on USB power: non need to warn.

#define       DEBOUNCE_TIME     20                              // Delay to debounce buttons, 20ms.
#define       LONG_PRESS        1000                            // Long button press, 1000ms.
#define       CYCLE_DELAY       100                             // Time to wait at the end of each cycle, 100ms.

LiquidCrystal lcd(8, 9, 4, 5, 6, 7);                            // Initialize the library with pin numbers and create "lcd" object (RS=pin8, EN=pin9, (R/W=gnd), D4=pin4, D5=pin5, D6=pin6, D7=pin7).
byte battery_char_1[8] =                                        // A charcater pattern for a nice paritally empty battery symbol.
{
    B00110,
    B01001,
    B01001,
    B01001,
    B01111,
    B01111,
    B01111,
};
byte battery_char_2[8] =                                        // A charcater pattern for a nice empty battery symbol.
{
    B00110,
    B01001,
    B01001,
    B01001,
    B01001,
    B01001,
    B01111,
};



// --------------------------------------------------------------------------
// This is the setup procedure executed once at startup. This procedure is
// automatically executed by Arduino.
// --------------------------------------------------------------------------
void setup()
{

    // Configure I/O lines.

    pinMode(IN_TUBE_1, INPUT);                                  // Configure tubes pulses pins as inputs.
    pinMode(IN_TUBE_2, INPUT);
    pinMode(IN_P1, INPUT_PULLUP);                               // Configure buttons and jumpers as inputs with pullup resistor.
    pinMode(IN_P2, INPUT_PULLUP);
    pinMode(IN_P3, INPUT_PULLUP);
    pinMode(IN_SINGLE_TUBE, INPUT_PULLUP);
    if (digitalRead(IN_SINGLE_TUBE) == LOW)                     // If D19/A5 is grounded, we have a single tube.
        num_tubes = 1;

    attachInterrupt(0, int_tube_1, FALLING);                    // Enable independent interrupts on change for D2 and D3.
    attachInterrupt(1, int_tube_2, FALLING);

    analogReference(INTERNAL);                                  // Use the internal 1.1V analog reference.
    pinMode(VOLT_VDD, INPUT);                                   // Voltage input on A0(D14) to read VDD voltage.
    pinMode(VOLT_BATT, INPUT);                                  // Voltage input on A1(D15) to read battery voltage.



    // Configure and initialize LCD display.

    lcd.begin(16, 2);                                           // Set up the LCD's display (16 columns by 2 rows).
    lcd.createChar(0, battery_char_1);                          // Set character 250 as the paritally empty battery character.
    lcd.createChar(1, battery_char_2);                          // Set character 251 as the empty battery character.



    // Configure serial port.

    Serial.begin(9600);                                         // Set serial communication at 9600N81.



    // Now we are done with initializations, we can show the welocme message.

    display_welcome_screen();                                   // Start with the usual greetings.
    
    display_dose(cpm_current);                                  // Display the current measure (zero, nothing was measurd yet, jut to "paint" the screen).
    next_millis = millis() + time_interval;                     // Set duration for first measurement.
}



// --------------------------------------------------------------------------
// This is the main program. It's executed automatically after the 
// initialization in an andless loop by Arduino.
// --------------------------------------------------------------------------
void loop()
{
    long int count_tube_1_tmp;                                  // Temporary variables to store current readings.
    long int count_tube_2_tmp;
    int volt_vdd;
    int volt_batt;

    lcd.setCursor(11, 0);                                       // Display the remaining time of the current measurement.
    display_time_left(long(next_millis-millis()), (time_interval <= 10000));   // Display tenths of a second if period is 10s or shorter.

    volt_vdd = read_voltage(VOLT_VDD);                          // Read Vdd rail and battery voltages.
    volt_batt = read_voltage(VOLT_BATT);

    if (millis() >= next_millis)                                // Check if integration time is elapsed.
    {
        next_millis = millis() + time_interval;                 // Yes, measurement is done: compute next interval.
        count_tube_1_tmp = count_tube_1;
        count_tube_2_tmp = count_tube_2;
        count_tube_1 = 0;                                       // Reset real-time counters.
        count_tube_2 = 0;

        cpm_current = count_tube_1_tmp + count_tube_2_tmp;

        cpm_current *= 60;                                      // Compute counts per minute.
        cpm_current /= (time_interval / 1000);

        cpm_current /= num_tubes;                               // Divide by 2 if using two tubes.
        cpm_current = compensate_dead_time(cpm_current);        // Compensate for tube dead time.

        compute_statistics(count_tube_1_tmp, count_tube_2_tmp); // Update MAX, min and Avg.

        lcd.setCursor(0, 0);                                    // Display result.
        display_dose(cpm_current);

        Serial.print(time_interval);                            // Communicate result on the serial port.
        Serial.print(';');
        Serial.print(count_tube_1_tmp);
        Serial.print(';');
        Serial.print(count_tube_2_tmp);
        Serial.print(';');
        Serial.print(cpm_current);
        Serial.print(';');
        Serial.println(volt_vdd);
    }

    lcd.setCursor(0, 1);                                        // Select second line and display values acoording to the selected mode.
    switch (mode)
    {
        case MODE_COOUNT    : lcd.print(count_tube_1);
                              lcd.print(" ");
                              lcd.print(count_tube_2);
                              lcd.print("             ");
                              break;
        case MODE_MAX       : display_dose(cpm_max);
                              lcd.print(" MAX");
                              break;
        case MODE_AVG       : display_dose(cpm_avg);
                              lcd.print(" Avg");
                              break;
        case MODE_MIN       : display_dose(cpm_min);
                              lcd.print(" min");
                              break;
        case MODE_VOLT      : display_voltage(volt_vdd);
                              lcd.print("dd ");
                              display_voltage(volt_batt);
                              lcd.print("b");
                              break;
    }

    if (mode != MODE_VOLT)                                      // Check battery and warn if necessary (but noth when already displaying voltages).
    {
        lcd.setCursor(15, 1);
        check_battery(volt_vdd, volt_batt);
    }

    button_handler();                                           // Handle buttons.
    delay(CYCLE_DELAY);                                         // Wait some time and cycle again (and again).
}



// --------------------------------------------------------------------------
// This procedure compensate for the tube dead time. When a radioactive 
// particle strikes the tube, the gas is ionized and a pulse is generated. 
// Then the ionization disappears and the tube is ready again for the next 
// particle, but as long as it’s ionized, it cannot detect another particle. 
// This is called dead time and de facto shortens the integration time. If 
// "f" is the measured pulse frequency and "tau" is the tube dead time, the 
// corrected pulse frequency "f'" can be expressed as follows:
//
// f' = f / (1 - f * tau)
//
// But time units for "f" and "tau" must match. Here "f" is in cpm and "tau" 
// (TUBE_DEAD_TIME) in microseconds, so some conversion has to take place. 
// Furthermore, in order to use integer math functions, all values are 
// multiplied by 1000. 
// The SMB-20 has a dead time of 190us and a maximum frequency of 220'000cpm, 
// so with a factor of 1000, we get sufficient accuracy while still fitting 
// into long int varaibles.
// The correction increases as the count increases. For the SBM-20 tube, the 
// correction is about 1% at 3'000cpm, rises at 10% at 30'000cpm and is 50% 
// at 160'000cpm.
// --------------------------------------------------------------------------
long int compensate_dead_time(long int measured_cpm)
{
    measured_cpm = 1000L * measured_cpm / (1000L - measured_cpm * TUBE_DEAD_TIME / (1000L * 60L));
    return (measured_cpm);
}



// --------------------------------------------------------------------------
// This procedure will update MAX, min and Avg values. It's executed at the 
// end of each measurement. It needs individual tube counts as input, all the
// rest are global variables.
// --------------------------------------------------------------------------
void compute_statistics(long int count_1, long int count_2)
{
    if (cpm_current > cpm_max)                              // Update MAX and min.
        cpm_max = cpm_current;
    if ((cpm_current < cpm_min) || (cpm_min < 0))
        cpm_min = cpm_current;

    if (avg_status == AVG_RESET)                            // Compute average: first, if no average is compted yet, start it.
    {
        avg_total_count = count_1 + count_2;                // Start counting pulses and milliseconds.
        avg_total_millis = time_interval;
        cpm_avg = cpm_current;                              // Set it equal to the first measurement.
        avg_status = AVG_RUNNING;                           // And enable average computation.
    }
    else if (avg_status == AVG_RUNNING)                     // Normal average computation.
    {
        if (((avg_total_count + count_1 + count_2) < avg_total_count) ||
           ((avg_total_millis + time_interval) < avg_total_millis))  // Make sure counters willnot overflow.
        {
            avg_status = AVG_OVERFLOW;                      // Stop computing average if they do.
        }
        else                                                // No risk of overflow, compute the average.
        {
            avg_total_count += count_1 + count_2;           // Count pulses and milliseconds.
            avg_total_millis += time_interval;
            if (avg_total_count < 40000000)                 // To compute counts per minute, precision is affected if avg_total_count is small or large. The actual thresold is not critical.
            {
                cpm_avg = avg_total_count * 60;             // If avg_total_count is small, first multiply by 60, then divide by the seconds (best precision, but will overflow with large values).
                cpm_avg /= (avg_total_millis / 1000);
            }
            else
            {
                cpm_avg = avg_total_count;
                cpm_avg /= (avg_total_millis / 1000);       // If avg_total_count is large, first multiply by the seconds, then divide by 60 (less precide, but won't overflow).
                cpm_avg *= 60;
            }
            cpm_avg /= num_tubes;                           // Divide by 2 if using two tubes.
            cpm_avg = compensate_dead_time(cpm_avg);        // Compensate for tube dead time.
        }
    }
}



// --------------------------------------------------------------------------
// This procedure will print the specified amount of milliseconds in 
// millis_to_go on the LCD screen at current cursor position. Time will be 
// displayed in seconds (divided by 1000). If use_decimal is non-zero, 
// tenths will be displayed as well. Numbers will be formatted to fit in 
// 5 characters: "xxx s" or "x.x s". If millis_to_go is larger than 1000000 
// or 10000 respectively, chrcters will extend to the right.
// --------------------------------------------------------------------------
void display_time_left(long int millis_to_go, char use_decimal)
{
    if (millis_to_go < 0)                                       // Ignore negative numbers and treat as 0.
        millis_to_go = 0;

    if (use_decimal == 0)                                       // If no decimal point is required, just align 100s and 10s of seconds,.
    {
        if (millis_to_go < 100000)
            lcd.print(" ");
        if (millis_to_go < 10000)
            lcd.print(" ");
        lcd.print(millis_to_go / 1000);                         // Convert milliseconds to seconds and write on display.
    }
    else
    {
        lcd.print(millis_to_go / 1000);                         // No alignmen is equird here as usually this diplays less than 10 seconds, just write the single digit.
        lcd.print(".");
        lcd.print((millis_to_go % 100) / 10);                   // Calculate and write the tenth digit.
    }
    lcd.print(" s");                                            // And also display the unit after a space.
}



// --------------------------------------------------------------------------
// This procedure will display the measured dose on the LCD screen at the 
// current cursor position. According to the global variable "unit", the 
// measurement can be converted into nano-Sievert per hour, micro-Roentgen 
// per hour or counts per minute. To perform the conversions, the two 
// constants TUBE_SV_FACTOR and TUBE_R_FACTOR are used.
//
// Counts per minute:
// ------------------
// designed for a maximum of 300000 cpm, the result is always displayed as
// "xxxxxx cpm" and right justified.
//
// nano-Sievert per hour:
// ----------------------
// Designed for an SBM-20 tube with 175cpm at 1uSv/h, it's designed to 
// display between 1nSv/h and 1.7mSv/h. Results are dsiplayed as 
// "  x. nSv/h"
// " xx. nSv/h"
// "xxx. nSv/h"
// "x.xx uSv/h"
// "xx.x uSv/h"
// "xxx. uSv/h"
// "x.xx mSv/h"
//
// micro-Roentgen per hour:
// ------------------------
// Designed for an SBM-20 tube with 1530cpm at 1uSv/h, it's designed to 
// display between 1mR/h and 196mR/h. Results are dsiplayed as 
// "  x. uR/h "
// " xx. uR/h "
// "xxx. uR/h "
// "x.xx mR/h "
// "xx.x mR/h "
// "xxx. mR/h "
//
// To allow displaying of consisten minimum values, if cpm_to_display is 
// negative, only blank spaces are displayed. 
// --------------------------------------------------------------------------
void display_dose(long int cpm_to_display)
{
    if (cpm_to_display < 0)
    {
        switch (unit)
        {
            case UNIT_R   : lcd.print("---. uR/h "); break;
            case UNIT_SV  : lcd.print("---. uSv/h"); break;
            default       : lcd.print("------ cpm"); break;
        }
        return;
    }

    if (unit == UNIT_R)
    {
        unsigned long int micro_roentgen_hour = cpm_to_display * 1000;
        micro_roentgen_hour /= TUBE_R_FACTOR;

        if (micro_roentgen_hour < 1000)
        {
            if (micro_roentgen_hour < 100)
                lcd.print(" ");
            if (micro_roentgen_hour < 10)
                lcd.print(" ");
            lcd.print(micro_roentgen_hour);
            lcd.print(". u");
        }
        else if (micro_roentgen_hour < 10000)
        {
            lcd.print(micro_roentgen_hour / 1000);
            lcd.print(".");
            if ((micro_roentgen_hour % 1000) < 100)
                lcd.print("0");
            lcd.print((micro_roentgen_hour % 1000) / 10);
            lcd.print(" m");
        }
        else if (micro_roentgen_hour < 100000)
        {
            lcd.print(micro_roentgen_hour / 1000);
            lcd.print(".");
            lcd.print((micro_roentgen_hour % 1000) / 100);
            lcd.print(" m");
        }
        else
        {
            lcd.print(micro_roentgen_hour / 1000);
            lcd.print(". m");
        }
        lcd.print("R/h ");
    }

    else if (unit == UNIT_SV)
    {
        unsigned long int nano_sievert_hour = cpm_to_display * 1000;
        nano_sievert_hour /= TUBE_SV_FACTOR;

        if (nano_sievert_hour < 1000)
        {
            if (nano_sievert_hour < 100)
                lcd.print(" ");
            if (nano_sievert_hour < 10)
                lcd.print(" ");
            lcd.print(nano_sievert_hour);
            lcd.print(". n");
        }
        else if (nano_sievert_hour < 10000)
        {
            lcd.print(nano_sievert_hour / 1000);
            lcd.print(".");
            if ((nano_sievert_hour % 1000) < 100)
                lcd.print("0");
            lcd.print((nano_sievert_hour % 1000) / 10);
            lcd.print(" u");
        }
        else if (nano_sievert_hour < 100000)
        {
            lcd.print(nano_sievert_hour / 1000);
            lcd.print(".");
            lcd.print((nano_sievert_hour % 1000) / 100);
            lcd.print(" u");
        }
        else if (nano_sievert_hour < 1000000)
        {
            lcd.print(nano_sievert_hour / 1000);
            lcd.print(". u");
        }
        else
        {
            lcd.print(nano_sievert_hour / 1000000);
            lcd.print(".");
            if ((nano_sievert_hour % 1000000) < 100000)
                lcd.print("0");
            lcd.print((nano_sievert_hour % 1000000) / 10000);
            lcd.print(" m");
        }
        lcd.print("Sv/h");
    }

    else
    {
        if (cpm_to_display < 100000)
            lcd.print(" ");
        if (cpm_to_display < 10000)
            lcd.print(" ");
        if (cpm_to_display < 1000)
            lcd.print(" ");
        if (cpm_to_display < 100)
            lcd.print(" ");
        if (cpm_to_display < 10)
            lcd.print(" ");
    
        lcd.print(cpm_to_display);
    
        lcd.print(" cpm");
    }
}


// --------------------------------------------------------------------------
// This procedure will disply the welcome screen on the LCD display and also
// send it on the serial port.
// --------------------------------------------------------------------------
void display_welcome_screen(void)
{
    lcd.clear();                                                // Clear the screen (previous errors or messages could be currently on the display).
    lcd.print(WELCOME_MSG_1);                                   // Print first welcome line.
    lcd.setCursor(0, 1);                                        // Move to line 2.
    lcd.print(WELCOME_MSG_2);                                   // Print second welcome line.

    Serial.println(WELCOME_MSG_1);                              // Print welcome msage also on the serial port.
    Serial.println(WELCOME_MSG_2);
    Serial.println();

    delay(2000);                                                // Wait 2 seconds.

    lcd.clear();                                                // Clear the screen again.
    lcd.print(num_tubes);
    Serial.print(num_tubes);
    lcd.print("x SBM-20 tube");
    Serial.print("x SBM-20 tube");
    if (num_tubes != 1)
    {
        lcd.print("s");
        Serial.print("s");
    }
    lcd.setCursor(0, 1);
    Serial.println();
    lcd.print("T_off = ");
    Serial.print("T_off = ");
    lcd.print(TUBE_DEAD_TIME);
    Serial.print(TUBE_DEAD_TIME);
    lcd.print(" us");
    Serial.print(" us");
    Serial.println();
    delay(1000);                                                // Wait 1 second.

    lcd.clear();                                                // Clear the screen again.
    lcd.print("1 mR/h = ");
    lcd.print(TUBE_R_FACTOR);
    lcd.print(" cpm");
    Serial.print("1 mR/h = ");
    Serial.print(TUBE_R_FACTOR);
    Serial.println(" cpm");
    lcd.setCursor(0, 1);
    lcd.print("1 uSv/h = ");
    lcd.print(TUBE_SV_FACTOR);
    lcd.print(" cpm");
    Serial.print("1 uSv/h = ");
    Serial.print(TUBE_SV_FACTOR);
    Serial.println(" cpm");
    Serial.println();
    delay(1000);                                                // Wait 1 second.

    lcd.clear();
    Serial.println("Time period [ms];Count tube 1 [1];Count tube 2 [1];Calculated counts per minute [cpm];Vdd voltage [mV]");
}



// --------------------------------------------------------------------------
// Interrupt service routine (INT0) run each time tube 1 generates a pulse.
// --------------------------------------------------------------------------
void int_tube_1()
{
    count_tube_1++;
}



// --------------------------------------------------------------------------
// Interrupt service routine (INT1) run each time tube 2 generates a pulse.
// --------------------------------------------------------------------------
void int_tube_2()
{
    count_tube_2++;
}



// --------------------------------------------------------------------------
// This procedure will return the rail voltage by reading the stepped down 
// voltage on the specified analog pin.
// The voltage divider is composed by two resistors R1=47k and R2=10k, step 
// down factor is therefore 1:5.7. 
// The ADC converter in the Arduino has a resolution of ADC_M=1024 points 
// and a voltage reference U_ref=1.1V. Knowing the ADC raw reading ADC#, we 
// can calculate the rail voltage U_rail with the following relation:
//
// U_rail = ADC# * (U_ref / ADC_M) * ((R1 + R2) / R2)
//
// Voltage is returned in millivolts, but the resolution is only 
//
// resolution = (U_ref / ADC_M) * ((R1 + R2) / R2) = 6.1 mV
//
// It's important to use unsigned long int variables and specify constants 
// as "UL" to avoid rounding errors.
// --------------------------------------------------------------------------
unsigned int read_voltage(char pin)
{
    unsigned long adc_value;

    adc_value = analogRead(pin);                                // Read the voltage on VOLT_IN.

    adc_value = adc_value * 1100UL * (10UL + 47UL);             // Calculate rail voltage with the equation explained above.
    adc_value = adc_value / (1024UL * 10UL);
    return(adc_value);
}



// --------------------------------------------------------------------------
// This procedure will display on the LCD screen the specified voltage.
// The variable is expressed in millivolts and it will be written in the
// "x.xx V" form, unit included, at current cursor position. For voltages 
// larger than 9.99V (not used here), everything will be shifted to the 
// right.
// --------------------------------------------------------------------------
void display_voltage(unsigned int voltage_to_display)
{
    lcd.print(voltage_to_display / 1000);                       // Convert millivolts to volts and write on display.
    lcd.print(".");

    voltage_to_display = voltage_to_display % 1000;             // Get rid of the volts and keep only millivolts.
    voltage_to_display = voltage_to_display / 10;               // Not interested in millivolt units, just keep 10s and 100s millivolts.

    if (voltage_to_display < 10)                                // Add leading zeros for voltages smaller than 10mV.
        lcd.print("0");
    lcd.print(voltage_to_display);                              // Display decimal figures.

    lcd.print(" V");                                            // And also display the unit after a space.
}



// --------------------------------------------------------------------------
// This procedure will take care of the buttons and is executed once at each
// main loop.
// --------------------------------------------------------------------------
void button_handler(void)
{
    if (digitalRead(IN_P1) == LOW)                              // Check if button 1 is prssed and cycle between available units.
    {
        delay(DEBOUNCE_TIME);                                   // Debounce button.
        if (digitalRead(IN_P1) == LOW)                          // Check if it's still pressed (to avoid glitches).
        {
            while (digitalRead(IN_P1) == LOW) ;                 // Yes, still pressed, wait until it's released.
            switch (unit)                                       // Vaid button press: cycle between available units. 
            {
                case UNIT_CPM  : unit = UNIT_R; break;
                case UNIT_R    : unit = UNIT_SV; break;
                case UNIT_SV   : unit = UNIT_CPM; break;
                default        : unit = UNIT_CPM; break;
            }
            lcd.setCursor(0, 0);                                // Update display.
            display_dose(cpm_current);
        }
    }

    if (digitalRead(IN_P2) == LOW)                              // Check if button 2 is prssed and cycle between available modes.
    {
        delay(DEBOUNCE_TIME);                                   // Debounce button.
        if (digitalRead(IN_P2) == LOW)                          // Check if it's still pressed (to avoid glitches).
        {
            unsigned long int old_time = millis();              // Save current milliseconds count to measure how long the button is pressed.
            while (digitalRead(IN_P2) == LOW)                   // Button is still pressed, wait until it's released.
            {
                if (millis() > (old_time + LONG_PRESS))         // While waiting, check if it's a long button press.
                {
                    lcd.setCursor(0, 1);                        // Inform user that stats are reset when the button is held long enough.
                    lcd.print("Statistics reset");
                }
            }
            if (millis() > (old_time + LONG_PRESS))             // Was it a long button press.
            {
                cpm_max = 0;                                    // Yes, reset statistics.
                cpm_min = -1;
                cpm_avg = -1;
                avg_total_count = 0;
                avg_total_millis = 0;
                avg_status = AVG_RESET;
            }
            else
            {
                switch (mode)                                   // No, short press: just cycle between available modes. 
                {
                    case MODE_COOUNT   : mode = MODE_MAX; break;
                    case MODE_MAX      : mode = MODE_AVG; break;
                    case MODE_AVG      : mode = MODE_MIN; break;
                    case MODE_MIN      : mode = MODE_VOLT; break;
                    case MODE_VOLT     : mode = MODE_COOUNT; break;
                    default            : mode = MODE_COOUNT; break;
                }
            }
            lcd.setCursor(0, 1);                                // Update (clear) display.
            lcd.print("                ");
        }
    }

    if (digitalRead(IN_P3) == LOW)                              // Check if button 3 is prssed and cycle between available time inervals.
    {
        delay(DEBOUNCE_TIME);                                   // Debounce button.
        if (digitalRead(IN_P3) == LOW)                          // Check if it's still pressed (to avoid glitches).
        {
            while (digitalRead(IN_P3) == LOW) ;                 // Yes, still pressed, wait until it's released.
            switch (time_interval)                              // Vaid button press: cycle between available units. 
            {
                case TIME_SLOW : time_interval = TIME_MED; break;
                case TIME_MED  : time_interval = TIME_FAST; break;
                case TIME_FAST : time_interval = TIME_SLOW; break;
                default        : time_interval = TIME_MED; break;
            }
            next_millis = millis();                             // Reset timer so that a new measure will immediately start.
        }
    }
}



// --------------------------------------------------------------------------
// This procedure will compare the pased voltages and display a low battery 
// Symbol if the battery is low. Every 5 time this procedure is called, the
// symbol will change to make it blink between two different symbols.
// If Vdd is below 4.5V, we definitely have a voltage problem, so a low 
// battery warining is produced.
// The LiPo battery is conidered low if below 2.9V, but if the battery 
// voltage is below 0.1V, we have no battery and we are running on USB 
// power: non need to warn.
// --------------------------------------------------------------------------
void check_battery(int vdd, int vbatt)
{
    static byte blink_counter = 0;                              // Counter used to slow down the blinking of the empty battery.

    if ((vdd < VDD_MIN) || ((vbatt < VBATT_MIN) && (vbatt > VBATT_ZERO)))    // Check voltages.
    {
        if (blink_counter < 5)                                  // Cycle between the two empty battery symbols.
            lcd.print('\0');
        else
            lcd.print('\1');
        blink_counter++;
        if (blink_counter > 10)
            blink_counter = 0;
    }
    else
    {
        lcd.print(' ');                                         // Battery is ok, clear the character.
    }
}




